<?php

namespace Drupal\file_update;

use Drupal\Core\Entity\EntityMalformedException;
use Drupal\Core\Entity\EntityStorageException;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\File\FileSystem;
use Drupal\Core\Image\ImageFactory;
use Drupal\Core\Logger\LoggerChannelInterface;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\Session\AccountProxyInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\file\FileInterface;
use Drupal\file_update\Plugin\FileUpdate\FileUpdateManager;
use Drupal\Core\ProxyClass\File\MimeType\MimeTypeGuesser;

/**
 * Class FileUpdateService.
 */
class FileUpdateService {

  use StringTranslationTrait;

  /**
   * Drupal\Core\Session\AccountProxyInterface definition.
   *
   * @var \Drupal\Core\Session\AccountProxyInterface
   */
  protected $currentUser;

  /**
   * Drupal\Core\Entity\EntityTypeManagerInterface definition.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * Drupal\Core\Logger\LoggerChannelInterface definition.
   *
   * @var \Drupal\Core\Logger\LoggerChannelInterface
   */
  protected $loggerChannel;

  /**
   * Drupal\Core\Messenger\MessengerInterface definition.
   *
   * @var \Drupal\Core\Messenger\MessengerInterface
   */
  protected $messenger;

  /**
   * Drupal\Core\Image\ImageFactory definition.
   *
   * @var \Drupal\Core\Image\ImageFactory
   */
  protected $imageFactory;

  /**
   * Drupal\file\FileStorage definition.
   *
   * @var \Drupal\file\FileStorage
   */
  protected $fileStorage;

  /**
   * Drupal\Core\File\FileSystem definition.
   *
   * @var \Drupal\Core\File\FileSystem
   */
  protected $fileSystem;

  /**
   * Drupal\Core\Extension\ModuleHandler definition.
   *
   * @var \Drupal\Core\Extension\ModuleHandler
   */
  protected $moduleHandler;

  /**
   * Drupal\Core\ProxyClass\File\MimeType\MimeTypeGuesser definition.
   *
   * @var \Drupal\Core\ProxyClass\File\MimeType\MimeTypeGuesser
   */
  protected $mimeTypeGuesser;

  /**
   * Drupal\file_update\Plugin\FileUpdate\FileUpdateFieldManager definition.
   *
   * @var \Drupal\file_update\Plugin\FileUpdate\FileUpdateManager
   */
  protected $pluginManager;

  /**
   * File entity set on init service.
   *
   * @var \Drupal\file\FileInterface
   */
  private $fileEntity;

  /**
   * File properties set on init service.
   *
   * @var array
   */
  private $fileProps;

  /**
   * Drupal\file_update\Plugin\FileUpdate\FileUpdateInterface definition.
   *
   * @var \Drupal\file_update\Plugin\FileUpdate\FileUpdateInterface[]
   */
  protected $plugins;

  /**
   * Constructs a new FileUpdateService object.
   *
   * @param \Drupal\Core\Session\AccountProxyInterface $current_user
   *   Current user service.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   Entity type manager service.
   * @param \Drupal\Core\Logger\LoggerChannelInterface $logger_channel
   *   Logger channel service.
   * @param \Drupal\Core\Messenger\MessengerInterface $messenger
   *   Messenger service.
   * @param \Drupal\Core\Image\ImageFactory $image_factory
   *   Image factory service.
   * @param \Drupal\Core\File\FileSystem $file_system
   *   File system service.
   * @param \Drupal\file_update\Plugin\FileUpdate\FileUpdateManager $file_update_manager
   *   FileUpdate plugin manager.
   * @param \Drupal\Core\ProxyClass\File\MimeType\MimeTypeGuesser $mime_type_guesser
   *   Mime type guesser service.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function __construct(
    AccountProxyInterface $current_user,
    EntityTypeManagerInterface $entity_type_manager,
    LoggerChannelInterface $logger_channel,
    MessengerInterface $messenger,
    ImageFactory $image_factory,
    FileSystem $file_system,
    FileUpdateManager $file_update_manager,
    MimeTypeGuesser $mime_type_guesser
  ) {
    $this->currentUser = $current_user;
    $this->entityTypeManager = $entity_type_manager;
    $this->loggerChannel = $logger_channel;
    $this->messenger = $messenger;
    $this->imageFactory = $image_factory;
    $this->fileSystem = $file_system;
    $this->pluginManager = $file_update_manager;
    $this->mimeTypeGuesser = $mime_type_guesser;

    $this->fileStorage = $this->entityTypeManager->getStorage('file');
  }

  /**
   * Init serve for operations on specific file.
   *
   * @param \Drupal\file\FileInterface|int $file
   *   Current file entity to work on.
   *
   * @return bool
   *   Returns TRUE if operation was successful.
   */
  protected function initService($file) {
    if (!$this->currentUser->hasPermission('update file entities')) {
      return FALSE;
    }
    $file_entity = $this->getFile($file);
    if ($this->fileEntity
      && $file_entity instanceof FileInterface
      && $this->fileEntity->id() == $file_entity->id()
    ) {
      return TRUE;
    }
    elseif ($file_entity instanceof FileInterface) {
      $this->fileEntity = $file_entity;
      return $this->setFileProps() && $this->setPlugins();
    }
    else {
      $this->messenger->addWarning($this->t('FileUpdateService can not initialize.'));
      return FALSE;
    }
  }

  /**
   * Image file loader.
   *
   * @param int|FileInterface $file
   *   The file entity id.
   *
   * @return \Drupal\file\FileInterface|false
   *   Returns File entity or FALSE if $entity is not an image file entity.
   */
  protected function getFile($file) {
    if ($file instanceof FileInterface) {
      $file_entity = $file;
    }
    else {
      /** @var \Drupal\file\Entity\File $file_entity */
      $file_entity = $this->fileStorage->load($file);
      if (!$file_entity) {
        $this->messenger->addWarning($this->t('File id %id not exists.', ['%id' => $file]));
        return FALSE;
      }
    }
    return $file_entity;
  }

  /**
   * Get configured plugins.
   *
   * @return bool
   *   Returns TRUE if operation was successful.
   */
  protected function setPlugins() {
    try {
      $config = [
        'file_entity' => $this->fileEntity,
        'file' => $this->getFileProps($this->fileEntity),
      ];
      $collector = [];
      foreach ($this->pluginManager->getDefinitions() as $key => $plugin_def) {
        /** @var \Drupal\file_update\Plugin\FileUpdate\FileUpdateInterface $plugin */
        $collector[$key] = $this->pluginManager->createInstance($key, $config);
      }
      $this->plugins = $collector;
      if ($usage = $this->getFileUsage()) {
        $this->fileProps['usage'] = $usage;
      }
      return TRUE;
    }
    catch (\Exception $e) {
      $this->messenger->addError($e->getMessage());
    }
    return FALSE;
  }

  /**
   * Set file props to class attributes.
   *
   * @return bool
   *   Returns TRUE if operation was successful.
   */
  protected function setFileProps() {
    $data = [
      'id' => $this->fileEntity->id(),
      'uuid' => $this->fileEntity->uuid(),
      'name' => $this->fileEntity->getFilename(),
      'mime' => $this->fileEntity->getMimeType(),
      'uri' => $this->fileEntity->getFileUri(),
      'url' => file_create_url($this->fileEntity->getFileUri()),
      'usage' => '',
      'size' => format_size($this->fileEntity->getSize()),
      'created' => date('Y-m-d h:m', $this->fileEntity->getCreatedTime()),
    ];
    try {
      $data['owner'] = $this->fileEntity->getOwner()->toLink();
    }
    catch (EntityMalformedException $e) {
    }
    $image = $this->imageFactory->get($data['uri']);
    if ($image->isValid()) {
      $data['width'] = $image->getWidth();
      $data['height'] = $image->getHeight();
      $data['preview'] = [
        '#theme' => 'image_style',
        '#style_name' => 'medium',
        '#uri' => $data['uri'],
      ];
    }
    else {
      $data['preview'] = FALSE;
    }
    $this->fileProps = $data;
    return TRUE;
  }

  /**
   * Returns a collection of file props.
   *
   * @param \Drupal\file\FileInterface|int $file_entity
   *   The file entity or ID.
   *
   * @return array|false
   *   Array with file properties.
   */
  public function getFileProps($file_entity) {
    if ($this->initService($file_entity)) {
      return $this->fileProps;
    }
    return FALSE;
  }

  /**
   * Returns an overview of file usages that will be updated as a string.
   *
   * @return string
   *   Collected usage separated by comma.
   */
  protected function getFileUsage() {
    $collector = [];
    foreach ($this->plugins as $plugin) {
      $collector = array_merge($collector, $plugin->getUsage());
    }
    return implode(', ', $collector);
  }

  /**
   * Main function.
   *
   * @param \Drupal\file\FileInterface|int $file
   *   The old name of the image used as identifier.
   * @param array $options
   *   The alt text of the image.
   *   - move_to: string
   *       update filename, path/, path/filename or schemed uri for SEO.
   *   - validate_max_size: int 0 or 1
   *       update images larger then setting.
   *   - max_width: int
   *       the max resolution width.
   *   - max_height: int
   *       the max resolution height.
   *
   * @return bool
   *   Returns TRUE if operation was successful.
   */
  public function updateFile($file, array $options) {
    if ($this->initService($file)) {
      // Move file.
      if (isset($options['move_to']) && $options['move_to']) {
        $this->moveFile($options['move_to']);
      }

      // Resize physical file.
      if (isset($options['validate_max_size'])) {
        $this->fileValidateMaxResolution($options['validate_max_size']);
      }
      return TRUE;
    }
    return FALSE;
  }

  /**
   * Validates max resolution if file is an image.
   *
   * @param string $max_res
   *   Can be an empty array. You can set specific
   *   'max_with' and 'max_height' here.
   *
   * @return bool
   *   If operation was successful.
   */
  protected function fileValidateMaxResolution($max_res) {
    if ($this->isValidMaxRes($max_res)) {
      $results = file_validate_image_resolution($this->fileEntity, $max_res);
      if (!count($results)) {
        try {
          if ($this->fileEntity->save() == SAVED_UPDATED) {
            $image = $this->imageFactory->get($this->fileEntity->getFileUri());
            $width = $image->getWidth();
            $height = $image->getHeight();

            foreach ($this->plugins as $plugin) {
              $plugin->updateResolution($width, $height);
            }
          }
        }
        catch (EntityStorageException $e) {
          $this->messenger->addError($e->getMessage());
          $this->loggerChannel->error($e->getMessage());
        }
      }
      else {
        foreach ($results as $result) {
          $this->messenger->addError($result);
          $this->loggerChannel->error($result);
        }
      }
    }
    return TRUE;
  }

  /**
   * Checks if the value has correct form of {width}x{height}.
   *
   * Width and height must be >=10 and <=99999.
   *
   * @param string $value
   *   String to check.
   *
   * @return bool
   *   If string is a valid max resolution.
   */
  public function isValidMaxRes($value) {
    return 1 == preg_match('/^[0-9]{2,5}x[0-9]{2,5}$/', $value);
  }

  /**
   * Move or update file entity.
   *
   * @param string $filename
   *   A file name, path, path/file_name, uri.
   *
   * @return bool
   *   If operation was successful.
   */
  protected function moveFile($filename) {
    $uri_new = $this->prepareNewUri($this->fileEntity, $filename);
    $uri_old = $this->fileEntity->getFileUri();
    $fid = $this->fileEntity->id();
    // Check new filename differs from old.
    if ($uri_new != $uri_old) {
      $dir = $this->prepareDirectory($uri_new);

      if ($dir && $move_success = file_move($this->fileEntity, $uri_new)) {
        try {
          if ($move_success->save() == SAVED_UPDATED) {
            $this->fileEntity = $move_success;
            $this->messenger->addMessage($this->t('Moved image (%id): %old => %new', [
              '%id' => $fid,
              '%old' => $uri_old,
              '%new' => $uri_new,
            ]));

            foreach ($this->plugins as $plugin) {
              $plugin->updateUri($uri_new);
            }
            return TRUE;
          }
        }
        catch (\Exception $exception) {
          $this->messenger->addError($exception->getMessage());
          $this->loggerChannel->error($exception->getMessage());
        }
      }
      else {
        $message = $this->t('Could not move image (%id): %old => %new', [
          '%id' => $fid,
          '%old' => $uri_old,
          '%new' => $uri_new,
        ]);
        $this->messenger->addWarning($message);
        $this->loggerChannel->warning('Could not move image (%id): %old => %new', [
          '%id' => $fid,
          '%old' => $uri_old,
          '%new' => $uri_new,
        ]);
      }
    }
    return FALSE;
  }

  /**
   * Prepares a new directory for a file uri if not exists or is not writeable.
   *
   * @param string $uri
   *   The complete uri including file name.
   *
   * @return bool
   *   Successful or not.
   */
  protected function prepareDirectory($uri) {
    $path = explode('/', $uri);
    /* $file_name = */ array_pop($path);
    $path = implode('/', $path);
    $dir = $this->fileSystem->prepareDirectory($path, FileSystem::CREATE_DIRECTORY);
    $write = $this->fileSystem->prepareDirectory($path, FileSystem::MODIFY_PERMISSIONS);
    return ($dir && $write);
  }

  /**
   * Create the new uri for moving file from various kinds of strings.
   *
   * @param \Drupal\file\FileInterface|int $file
   *   The file entity working on.
   * @param string $filename
   *   The filename e.g., path or uri.
   *
   * @return string|false
   *   The new uri.
   *   ''                                 -> public://old/path/old_file_name.jpg
   *   new_file_name                      -> public://old/path/new_file_name.jpg
   *   new_file_name.jpg                  -> public://old/path/new_file_name.jpg
   *   new_file_name.gif                  -> public://old/path/new_file_name.jpg
   *   new/path/new_file_name.jpg         -> public://new/path/new_file_name.jpg
   *   new/path/                          -> public://new/path/old_file_name.jpg
   *   /new/path/                         -> public://new/path/old_file_name.jpg
   *   public://folder/new_file_name.jpg  -> public://folder/my_super_file.jpg
   *   private://folder/new_file_name.jpg -> private://folder/my_super_file.jpg
   */
  public function prepareNewUri($file, $filename = NULL) {
    /** @var \Drupal\file\FileInterface $file_entity */
    $file_entity = $this->getFile($file);
    if (!$file_entity) {
      return FALSE;
    }

    if (preg_match('/\.[a-zA-Z0-9]{2,7}$/', $filename) == 1) {
      // It's not just a path that ends with a "/".
      $mime_guess = $this->mimeTypeGuesser->guess($filename);
      if ($mime_guess == $file_entity->getMimeType()) {
        $new_filename_frags = explode('.', $filename);
        $ending = array_pop($new_filename_frags);
      }
      else {
        $this->messenger->addError($this->t('File has not correct extension.'));
        return FALSE;
      }
    }

    if (!isset($ending)) {
      $old_path_frags = explode('.', $file_entity->getFileUri());
      $ending = array_pop($old_path_frags);
    }
    $ending = ".{$ending}";

    $new_path_frags = explode('/', $filename);
    $old_uri = $file_entity->getFileUri();
    // Analyse filename.
    if (!$filename) {
      // No file name is given.
      return $old_uri;
    }
    elseif (count($new_path_frags) == 1) {
      // Just a file name given without path. Use old path with new file name.
      $filename_frags = explode('.', $filename);
      $filename = reset($filename_frags);
      $frags = explode('/', $old_uri);
      // Remove original filename.
      array_pop($frags);
      $frags[] = $filename . $ending;
      return implode('/', $frags);
    }
    elseif (
      count($new_path_frags) > 2
    // Means double slash.
      && $new_path_frags[1] == ''
    // Means stream wrapper e.g. "public:".
      && 1 == preg_match('/^[a-z]{2,12}:$/', $new_path_frags[0])
    ) {
      // URI given with stream wrapper and optional path.
      $filename = array_pop($new_path_frags);
      if ($filename == '') {
        // Just a new path given, use old file name.
        $new_path_frags[] = $file_entity->getFilename();
      }
      else {
        $filename_frags = explode('.', $filename);
        $filename = reset($filename_frags);
        $new_path_frags[] = $filename . $ending;
      }
      return implode('/', $new_path_frags);
    }
    else {
      // Path given without stream wrapper.
      $filename = array_pop($new_path_frags);
      // Remove empty to avoid trailing slashes.
      $new_path_frags = array_filter($new_path_frags);
      if ($filename == '') {
        // Just a new path given, use old file name.
        $new_path_frags[] = $file_entity->getFilename();
      }
      else {
        $filename_frags = explode('.', $filename);
        $filename = reset($filename_frags);
        $new_path_frags[] = $filename . $ending;
      }
      $new_path = implode('/', $new_path_frags);
      return 'public://' . $new_path;
    }
  }

  /**
   * Check if uri is valid.
   *
   * @return bool
   *   If uri is valid.
   */
  public function isValidUri($uri) {
    return $this->fileSystem->validScheme($uri);
  }

}
