<?php

namespace Drupal\file_update\Plugin\FileUpdate;

use Drupal\Core\Entity\EntityTypeManager;
use Drupal\Core\Entity\FieldableEntityInterface;
use Drupal\Core\Extension\ModuleHandler;
use Drupal\Core\Logger\LoggerChannel;
use Drupal\Core\Messenger\Messenger;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\file\FileUsage\DatabaseFileUsageBackend;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Plugin implementation file_update plugins.
 *
 * @FileUpdate(
 *   id = "editor",
 *   label = @Translation("File reference from editor fields."),
 *   description = @Translation("Updates fields of type editor (core)."),
 *   permission = "administer nodes",
 * )
 */
class FileUpdateEditor extends FileUpdateBase implements FileUpdateInterface, ContainerFactoryPluginInterface {

  /**
   * Drupal\Core\Entity\EntityTypeManager definition.
   *
   * @var \Drupal\Core\Entity\EntityTypeManager
   */
  protected $entityTypeManager;

  /**
   * Drupal\Core\Extension\ModuleHandler definition.
   *
   * @var \Drupal\Core\Extension\ModuleHandler
   */
  protected $moduleHandler;

  /**
   * Drupal\Core\Messenger\MessengerInterface definition.
   *
   * @var \Drupal\Core\Messenger\MessengerInterface
   */
  protected $messenger;

  /**
   * Drupal\Core\Logger\LoggerChannelInterface definition.
   *
   * @var \Drupal\Core\Logger\LoggerChannelInterface
   */
  protected $loggerChannel;

  /**
   * Drupal\file\FileUsage\DatabaseFileUsageBackend definition.
   *
   * @var \Drupal\file\FileUsage\DatabaseFileUsageBackend
   */
  protected $fileUsage;

  /**
   * {@inheritdoc}
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition, EntityTypeManager $entity_type_manager, ModuleHandler $module_handler, LoggerChannel $logger_channel, Messenger $messenger, DatabaseFileUsageBackend $file_usage) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);
    $this->entityTypeManager = $entity_type_manager;
    $this->moduleHandler = $module_handler;
    $this->loggerChannel = $logger_channel;
    $this->messenger = $messenger;
    $this->fileUsage = $file_usage;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('entity_type.manager'),
      $container->get('module_handler'),
      $container->get('logger.channel.file_update'),
      $container->get('messenger'),
      $container->get('file.usage')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function isRequired() {
    return TRUE;
  }

  /**
   * {@inheritdoc}
   */
  public function getUsage() {
    $collector = [];
    foreach ($this->getReferrer() as $key => $usage) {
      foreach ($usage as $entity_type => $entities) {
        foreach ($entities as $entity_id => $entity) {
          $collector[] = "$entity_type:$entity_id->$key";
        }
      }
    }
    return $collector;
  }

  /**
   * {@inheritdoc}
   */
  public function updateUri($uri) {
    $uuid = $this->getFileEntity()->uuid();
    $url = file_url_transform_relative(file_create_url($uri));
    $pattern = '/<[^>]*' . preg_quote($uuid) . '[^>]*>/';
    try {
      foreach ($this->getReferrer() as $field_name => $usage) {
        foreach ($usage as $entity_type => $entity_ids) {
          $entities = $this->entityTypeManager->getStorage($entity_type)
            ->loadMultiple($entity_ids);
          foreach ($entities as $entity_id => $entity) {
            if ($entity instanceof FieldableEntityInterface) {
              $values = $entity->get($field_name)->getValue();
              foreach ($values as $key => $value) {
                $edit_str = $value['value'];
                preg_match($pattern, $value['value'], $matches);
                foreach ($matches as $match) {
                  $new_value = preg_replace('/src="([^"]*)"/', "src=\"{$url}\"", $match);
                  $edit_str = str_replace($match, $new_value, $edit_str);
                }
                $values[$key]['value'] = $edit_str;
              }
              $entity->get($field_name)->setValue($values);
              if ($entity->save() == SAVED_UPDATED) {
                $message = $this->t('Updated %type:%id entity at field %field with new image url "%url".', [
                  '%type' => $entity_type,
                  '%id' => $entity_id,
                  '%field' => $field_name,
                  '%url' => $url,
                ]);
                $this->messenger->addMessage($message);
                $this->loggerChannel->info('Updated %type:%id entity at field %field with new image url "%url".', [
                  '%type' => $entity_type,
                  '%id' => $entity_id,
                  '%field' => $field_name,
                  '%url' => $url,
                ]);
              }
            }
          }
        }
      }
    }
    catch (\Exception $e) {
      $this->messenger->addError($e->getMessage());
      $this->loggerChannel->error($e->getMessage());
    }
  }

  /**
   * {@inheritdoc}
   */
  public function updateResolution($width, $height) {
    return TRUE;
  }

  /**
   * Returns all crop entities using a file.
   *
   * @return array
   *   Array with field items referring to current file entity.
   */
  protected function getReferrer() {
    $collector = [];
    try {
      if ($this->isRequired()) {
        $uuid = $this->getFileEntity()->uuid();
        foreach ($this->fileUsage->listUsage($this->getFileEntity()) as $key => $usage) {
          if ($key != 'editor') {
            continue;
          }
          foreach ($usage as $entity_type => $references) {
            $entities = $this->entityTypeManager->getStorage($entity_type)
              ->loadMultiple(array_keys($references));
            foreach ($entities as $entity) {
              if ($entity instanceof FieldableEntityInterface) {
                foreach ($entity->getFieldDefinitions() as $field_name => $field_definition) {
                  $field_types = ['text', 'text_long', 'text_with_summary'];
                  if (in_array($field_definition->getType(), $field_types)) {
                    if (strpos($entity->get($field_name)->getString(), $uuid) >= 0) {
                      $collector[$field_name][$entity_type][$entity->id()] = $entity->id();
                    }
                  }
                }
              }
            }
          }
        }
      }
    }
    catch (\Exception $e) {
      $this->messenger->addError($e->getMessage());
      $this->loggerChannel->error($e->getMessage());
    }
    return $collector;
  }

}
